


/****************************************************************************
;
; NAME: osutil.c
;
;
; DESCRIPTION:
;
;       Linux utility wrap functions
;
;
; REVISION HISTORY:
;
;       $Log: /ClrtyFC8/Linux/driver/osutil.c $
; 
; 79    6/20/11 1:57p Pcullen
; Fix for IR7505 and associated kernel panics during driver removal.
; Also, instrumented Linux code with attolog statements.
; 
; 78    5/23/11 3:54p Jseba
; Fixes for 2.6.39
; 
; 77    4/22/11 8:52a Pcullen
; Changes for ESXi 5 RC celerity8fc driver; includes fix for IR7577.
; 
; 76    2/14/11 2:38p Pcullen
; Add support for NPIV in Celerity8 driver.
; 
; 75    9/08/10 9:36a Jseba
; Replaced LinuxGetSgDma with LinuxGetPhysAddrFromSgc, which includes
; support for chained scatter/gather lists; added LinuxGetRegionSize;
; modified LinuxWaitOn to poll the conditional in case
; wait_event_interruptible() returns an error
; 
; 74    8/09/10 4:38p Mseier
; sas_remove_host function is only used if kernel version is 2.6.18 or
; higher, instead of 2.6.14 or higher.  This resolves a conflict where a
; kernel between 2.6.14 and 2.6.17 would call sas_remove_host but not
; include the linux kernel file where it is defined (this happens in SUSE
; 10.2).
; 
; 73    6/15/10 10:05a Mseier
; Updated ESASRAID VDA events to use new async event naming scheme.
; 
; 72    6/01/10 12:54p Mseier
; Added support for slot binding and bad block reporting
; 
; 71    5/27/10 4:20p Pcullen
; Esas2hba port to VMware 4.x; added LinuxCmdGetDumpActive function.
; 
; 70    5/10/10 1:30p Pcullen
; Fixed conditional compilation on transport usage so that older (RH4)
; kernel don't attempt to use it.
; 
; 69    4/22/10 8:16a Pcullen
; Implementation of Linux SAS Transport
; 
; 68    4/14/10 1:31p Mseier
; Updates for ESASRAID2 project.
; 
; 67    3/01/10 2:31p Pcullen
; Updates for Celerity 8Gb Multipathing: Added support for linux
; completion API; Added LinuxScsiDone as common completion point for all
; scsi commands; Updated BH work API for new kernels and generalized it
; for more than just rescanning host - MP needs BH to process dev
; changes.
; 
; 66    12/17/09 3:28p Jseba
; Define use_transport_layer only if ATTO_USE_TRANSPORT is also defined
; 
; 65    11/20/09 4:19p Pcullen
; Rudimentary Linux SAS Transport implementation
; 
; 64    10/30/09 9:23a Pcullen
; VMware 4 porting
; Added inclusion of oscommon.h for VMware4 builds.
; Removed legacy ESX 3.5 changes.
; 
; 63    10/12/09 3:31p Jseba
; Added LinuxInRecovery to detect when the SCSI error handler is pending
; or running
; 
; 62    10/05/09 5:01p Jseba
; Removed all ATTO-specific data types
; 
; 61    8/06/09 1:29p Pcullen
; Provide more complete implementation of Linux FC Transport, including
; remote port api (rports)
; 
; 60    5/11/09 2:14p Jseba
; Add error values to SCSICMD_OFFSETS; change LinuxCompleteRequest to use
; error value from core instead of translating through the CompStatErr
; table
; 
; 59    3/18/09 8:20a Jseba
; Fixes for 2.6.24
; 
; 58    2/26/09 8:15a Jseba
; Fixed LinuxPciMap/UnmapSingle for >= 2.6.25
; 
; 57    2/11/09 9:48a Jseba
; Use a global, lockable scratch buffer instead of allocating large
; buffers on the stack when needed; minor formatting cleanups
; 
; 56    2/05/09 12:34p Pcullen
; Initial checkin from VMware 3.5 porting
; 
; 55    11/05/08 2:44p Jseba
; Make sure LinuxGetReqBuffer is properly kmapping the buffer, and add
; LinuxReleaseReqBuffer
; 
; 54    10/20/08 10:29a Jseba
; Fix compilation error on kernels 2.6.22 and older
; 
; 53    10/16/08 9:04a Jseba
; Added LinuxPciFindDevice, LinuxPciReleaseDevice, and LinuxGetBridgePdev
; 
; 52    10/07/08 4:50p Jseba
; Fix LinuxGetReqBuffer for 2.6.25
; 
; 51    9/30/08 10:14a Jseba
; Fix a bug in LinuxRescanHostRoutine that could lead to invalid
; outstanding references to the driver
; 
; 50    9/17/08 2:36p Jseba
; Add LinuxSscanf and LinuxGetReqBuffer; remove OsDebugHeader calls
; 
; 49    8/14/08 3:59p Jseba
; Add SCSICMD_FLAGS_USE_CMD_PTR for 2.6.26 
; 
; 48    7/14/08 3:22p Jseba
; Null timer pointer when killing it; change LinuxGetCurrentTimeMs to use
; do_gettimeofday to avoid jiffie wrap-around problems
; 
; 47    5/27/08 4:22p Jseba
; Add OsDebugPrint so wrapper can pass debug information to the core; add
; a scsicmd_off flag to specify whether the sense_buffer is a pointer 
; 
; 46    5/21/08 4:03p Jseba
; Add pci_set_consistent_dma_mask so Celerity can allocate memory over
; the 4GB barrier
; 
; 45    4/24/08 11:37a Jseba
; Fixes for kernel 2.6.25; change LinuxPciUnmapSg
; 
; 44    2/25/08 11:56a Jseba
; Removed LinuxGetSgDmaAddr
; 
; 43    2/19/08 10:58a Jseba
; Remove 64-bit pci_set_consistent_dma_mask call - we don't want
; consistent memory of the 4GB barrier
; 
; 42    1/23/08 2:15p Jseba
; Fix builds when ATTO_RESCANS_ENABLED is not defined
; 
; 41    1/08/08 9:19a Jseba
; Fix builds for older kernels
; 
; 40    1/07/08 3:45p Jseba
; Fix ESASRAID for 2.4 kernels
; 
; 39    11/27/07 5:05p Jseba
; Rework LinuxRescanHostRoutine to only add devices that do not already
; exist; remove atto_allow_rescans
; 
; 38    11/09/07 3:33p Jseba
; Fix LinuxDivideQwords for 2.4 builds
; 
; 37    11/06/07 12:58p Jseba
; Rework rescan code (change_notification=1) to fix problems with older
; kernels
; 
; 36    10/23/07 9:49a Jseba
; Don't call scsi_scan_target if SCAN_WILD_CARD is not defined
; 
; 35    10/17/07 4:13p Jseba
; Fixed rescan code for 2.6.9 and earlier
; 
; 34    10/16/07 10:01a Jseba
; Instead of using scsi_scan_host to rescan for new devices, use
; scsi_remove_device when devices are removed, and scsi_scan_target when
; devices are added
; 
; 33    9/21/07 3:11p Jseba
; Cleaned up print statements
; 
; 32    8/27/07 11:09a Jseba
; Use new flags for request_irq on newer kernel
; 
; 31    8/21/07 10:41a Jseba
; Changed LinuxRemoveDevice/LinuxAddDevice to LinuxRescanHost and
; implemented it
; 
; 30    8/06/07 3:27p Jseba
; Simplified LinuxCompleteRequest and LinuxDpcInit; use
; tasklet_hi_schedule for DPCs; removed LinuxPciMapSg (it is now done by
; this layer, not the core); added LinuxGetCurrentTimeUs and
; LinuxDivideQwords; added LinuxExchangeOffsets to allow core to directly
; access elements of command and SGL structures
; 
; 29    2/01/07 5:01p Jseba
; Added a workaround for a SCSI Tape driver bug in 2.4 kernels on x86_64
; platforms in LinuxPciMapSg/LinuxPciUnmapSg.
; 
; 28    11/08/06 11:36a Jseba
; Fix interruptible_sleep_on problem - kernel needs to be locked; move
; VDA event notification socket code to attocfg
; 
; 27    10/09/06 10:18a Jseba
; Use LinuxGetErrno for all drivers, not just ESASRAID
; 
; 26    10/09/06 9:49a Jseba
; Fix the waitqueue functions to pre-allocate and initialize the queue,
; and gave them more accurate names;  added LinuxGetErrno to translate
; ATTO error codes to Linux errno
; 
; 25    10/04/06 10:25a Jseba
; Restore LinuxMemMap/LinuxMemUnmap for use by celeritymem driver
; 
; 24    9/27/06 9:58a Jseba
; Remove unused LinuxCompareExchangeDword/Pointer
; 
; 23    9/22/06 5:21p Jseba
; Move pci_set_drvdata call from LinuxUnregisterScsiHost to new
; LinuxUnregisterPci; don't call cond_resched() if in interrupt context
; 
; 22    9/20/06 10:20a Jseba
; Move CompStatErr definitions to os.h; removed locking from
; LinuxCompleteRequest and added LinuxLockIo/UnlockIo for core to handle
; locking itself; removed LinuxClaimRegion and LinuxMemMap stuff,
; replaced with LinuxMapRegions/UnMapRegions
; 
; 21    9/14/06 11:36a Jseba
; ESASRAID async event support
; 
; 20    8/28/06 11:16a Jseba
; Support for ESASRAID driver; New LinuxMemMap/LinuxMemUnmap
; 
; 19    7/24/06 4:11p Jseba
; Conditionalize cond_resched call for 2.6 kernel only
; 
; 18    1/03/06 10:22a Jseba
; Remove deprecated "intermodule" functions; use conditional in
; LinuxSleepOn; fix bugs when initiator_mode==0 or !CONFIG_SMP; add
; LinuxYield
; 
; 17    8/04/05 11:52a Jseba
; Conditional out spinlock routines on uniprocessor systems; add
; LinuxGetKernelVersion.
; 
; 16    3/21/05 10:28a Jseba
; Remove LinuxLockIo/LinuxUnlockIo - just do the locking directly when
; necessary
; 
; 15    1/25/05 10:48a Jseba
; Added translations for RS_SEL2 and RS_ADDR_ERR
; 
; 14    12/28/04 3:45p Jseba
; Modify LinuxMemMap/Unmap to allow old-style usage
; 
; 13    12/28/04 9:51a Jseba
; Improved PCI initialization functions; misc. cleanup
; 
; 12    12/09/04 12:28p Jseba
; Use jiffies instead of do_gettimeofday for LinuxGetCurrentTimeMs - the
; resolution is good enough, and the overhead is much lower.
; 
; 11    11/29/04 3:48p Sschmitz
; Fix type mismatch in debug statement.
; 
; 10    11/19/04 11:22a Jseba
; 64-bit support; sysfs support
; 
; 9     7/16/04 2:12p Jseba
; Fixes for non-SMP systems (where spinlocks are NOPs)
; 
; 8     6/29/04 2:58p Jseba
; Modification so this module can be shared with celeritymem driver.
; 
; 7     6/16/04 1:12p Jseba
; Comment out unused declarations
; 
; 6     6/16/04 11:41a Jseba
; Use tasklets for DPC; make disable/enable IRQ a NOP for 2.6 kernel
; 
; 5     3/31/04 3:54p Jseba
; Initial support for 2.6 kernel
; 
; 4     12/10/03 3:22p Jseba
; Add LinuxDisableIrq & LinuxEnableIrq
; 
; 3     12/09/03 9:52a Jseba
; Rename to celerityfc
; 
; 2     11/18/03 10:24a Jseba
; Improved LinuxCompleteRequest
; 
; 1     10/24/03 10:38a Jseba
; Initial entry
; 
; 
;
;
; THIS PROGRAM AND THE INFORMATION CONTAINED HEREIN IS THE PROPERTY OF
; ATTO TECHNOLOGY, INC. AND SHALL NOT BE REPRODUCED, COPIED, OR USED IN
; WHOLE OR IN PART OTHER THAN AS PROVIDED FOR IN THE LICENSE AGREEMENT
; PURSUANT TO WHICH IT WAS FURNISHED.
;
; COPYRIGHT (c) ATTO TECHNOLOGY, INC. 2003
; ALL RIGHTS RESERVED.
;
;***************************************************************************/

#ifdef __VMKLNX__
#include "oscommon.h"
#ifdef ATTO_VMK_50
#include "vmklinux_scsi.h"
#endif /* ATTO_VMK_50 */
#else
#include "os.h"
#endif
#include "oswrap.h"

#ifdef ATTO_USE_TRANSPORT
#include "ostransport.h"
#endif

int global_pci_dev_offset;

#if defined (_MP_) && (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,27))
extern void (*scsi_cmnd_timeout_fn)(struct scsi_cmnd *);
#endif


/* Allocate a global, lockable scratch buffer for routines */
/* that need some space to scribble large chunks of data.  */

#ifdef DEFINE_SPINLOCK
DEFINE_SPINLOCK (scratch_buffer_lock);
#else
static spinlock_t scratch_buffer_lock = SPIN_LOCK_UNLOCKED;
#endif
static char scratch_buffer [1024];

unsigned long LinuxLockIo (void)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
    unsigned long flags;
    spin_lock_irqsave (&io_request_lock, flags);
    return flags;
#else
    return 0;
#endif
}


void LinuxUnlockIo (unsigned long flags)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
    spin_unlock_irqrestore (&io_request_lock, flags);
#endif
}       

void OsDebugPrint (char *fmt, ...)
{
    unsigned long flags;
    va_list ap;

    va_start (ap, fmt);

    spin_lock_irqsave (&scratch_buffer_lock, flags);

    vsprintf (scratch_buffer, fmt, ap);

    OsDebugString (scratch_buffer);

    spin_unlock_irqrestore (&scratch_buffer_lock, flags);
}

void LinuxScsiDone(void *_cmd)
{
    Scsi_Cmnd *cmd = (Scsi_Cmnd *) _cmd;

#if defined (_MP_) && (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,27))
    /* Restore the Scsi Command timeout timer */
    if (cmd->eh_timeout.function)
        del_timer (&cmd->eh_timeout);

    cmd->eh_timeout.data = (unsigned long) cmd;
    cmd->eh_timeout.expires =  jiffies + cmd->timeout_per_command;
    cmd->eh_timeout.function = (void (*)(unsigned long)) scsi_cmnd_timeout_fn;

    add_timer (&cmd->eh_timeout);
#endif

    /* Complete the command to the OS.    */
    cmd->scsi_done (cmd);
}


void LinuxCompleteRequest (void *_cmd, int scsi_stat, int error, int resid)
{
    Scsi_Cmnd *cmd = (Scsi_Cmnd *) _cmd;

    cmd->result = ((error << 16) | (scsi_stat & STATUS_MASK));

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
    cmd->resid = resid;
#else
    scsi_set_resid (cmd, resid);
#endif

    LinuxScsiDone (cmd);
}


unsigned long LinuxGetPhysAddrFromSgc (void *_sglist, unsigned long long *addr)
{
    struct scatterlist **sglist = (struct scatterlist **) _sglist;
    unsigned long       len    = sg_dma_len (*sglist);

    (*addr) = sg_dma_address (*sglist);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24) || defined (VMKLNX_VMKSGARRAY_SUPPORTED)
    (*sglist) = sg_next (*sglist);
#else
    (*sglist)++;
#endif

    return len;
}

unsigned long LinuxGetPhysAddrFromSgel (void *_sgel, unsigned long long *addr)
{
    struct scatterlist **sgel = (struct scatterlist **) _sgel;
    unsigned long       len    = sg_dma_len (*sgel);

    (*addr) = sg_dma_address (*sgel);

    return len;
}

void LinuxGetNextSgel (void *_sgel)
{
    struct scatterlist **sgel = (struct scatterlist **) _sgel;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24) || defined (VMKLNX_VMKSGARRAY_SUPPORTED)
    (*sgel) = sg_next (*sgel);
#else
    (*sgel)++;
#endif

}


void LinuxStallExecution (u32 Delay)
{
    while (Delay)
    {
        if (Delay < 1000)
        {
            udelay (Delay);
            barrier ();
            return;
        }
        else
        {
            mdelay (1);
            barrier ();
            Delay -= 1000;
        }
    }
}


int LinuxSprintf (char *str, const char *format, ...)
{
    va_list ap;

    va_start (ap, format);

    return vsprintf (str, format, ap);
}


int LinuxVsprintf (char *str, const char *format, va_list ap)
{
    return vsprintf (str, format, ap);
}


int LinuxSscanf (char *str, const char *format, ...)
{
#ifdef __VMKLNX__
    /* Apparently there is no vsscanf in the vmklinux api? */
    return 0;
#else
    va_list ap;

    va_start (ap, format);

    return vsscanf (str, format, ap);
#endif
}


void LinuxPrint (char *fmt, ...)
{
    unsigned long flags;
    va_list ap;

    va_start (ap, fmt);

    spin_lock_irqsave (&scratch_buffer_lock, flags);

    vsprintf (scratch_buffer, fmt, ap);

    printk (scratch_buffer);

    spin_unlock_irqrestore (&scratch_buffer_lock, flags);
}

void LinuxLog (char *fmt, ...)
{
    unsigned long flags;
    va_list ap;

    va_start (ap, fmt);

    spin_lock_irqsave (&scratch_buffer_lock, flags);

    vsprintf (scratch_buffer, fmt, ap);

    ATTO_LOG (KERN_INFO, ATTO_LOG_CORE, "%s", scratch_buffer);

    spin_unlock_irqrestore (&scratch_buffer_lock, flags);
}

void LinuxDevLog (void *_host, char *fmt, ...)
{
    unsigned long flags;
    va_list ap;
    struct Scsi_Host *shost;

    if (_host == NULL)
        return;

    shost = (struct Scsi_Host *)_host;

    va_start (ap, fmt);

    spin_lock_irqsave (&scratch_buffer_lock, flags);

    vsprintf (scratch_buffer, fmt, ap);

    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_CORE, &(shost->shost_gendev),
              "%s", scratch_buffer);

    spin_unlock_irqrestore (&scratch_buffer_lock, flags);
}


int LinuxPciReadConfig (void *_pdev, int where, void *ptr, int size)
{
    struct pci_dev *pdev = (struct pci_dev *) _pdev;

    switch (size)
    {
    case 1: return pci_read_config_byte  (pdev, where, (u8 *)  ptr);
    case 2: return pci_read_config_word  (pdev, where, (u16 *) ptr);
    case 4: return pci_read_config_dword (pdev, where, (u32 *) ptr);
    }

    return 0;
}


int LinuxPciWriteConfig (void *_pdev, int where, u32 val, int size)
{
    struct pci_dev *pdev = (struct pci_dev *) _pdev;

    switch (size)
    {
    case 1: return pci_write_config_byte  (pdev, where, (u8)  val);
    case 2: return pci_write_config_word  (pdev, where, (u16) val);
    case 4: return pci_write_config_dword (pdev, where, (u32) val);
    }

    return 0;
}


void *LinuxPciAllocConsistent (void *_pdev, int size, unsigned long long *bus_addr)
{
    ATTO_LOG (KERN_INFO, ATTO_LOG_PCI,
              "pci_alloc_consistent called; size:%d bus_addr:%p\n",
              size, bus_addr);

    return pci_alloc_consistent ((struct pci_dev *) _pdev,
                                 (size_t) size,
                                 (dma_addr_t *) bus_addr);
}


void LinuxPciFreeConsistent (void *_pdev, int size, void *cpu_addr, unsigned long long bus_addr)
{
    pci_free_consistent ((struct pci_dev *) _pdev,
                         (size_t) size, cpu_addr,
                         (dma_addr_t) bus_addr);
}


void LinuxUnMapRegions (void *_pdev, int *bars, int num_bars, void **maps)
{
    struct pci_dev *pdev = (struct pci_dev *) _pdev;
    int cur_bar;

    for (cur_bar = 0; cur_bar < num_bars; cur_bar++)
    {
        if (maps [cur_bar])
            iounmap (maps [cur_bar]);

        maps [cur_bar] = NULL;

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
        pci_release_region (pdev, bars [cur_bar]);
#endif
    }

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    pci_release_regions (pdev);
#endif
}


int LinuxMapRegions (void *_pdev, char *name, int *bars, int num_bars, void **maps)
{
    struct pci_dev *pdev = (struct pci_dev *) _pdev;
    int cur_bar;

    /* First make sure the maps are cleared */

    for (cur_bar = 0; cur_bar < num_bars; cur_bar++)
        maps [cur_bar] = NULL;

    /* Reserve all requested regions */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    if (pci_request_regions (pdev, name))
    {
        printk (DEBUG_HEADER "Unable to reserve MMIO regions.\n");

        return 0;
    }
#else
    for (cur_bar = 0; cur_bar < num_bars; cur_bar++)
    {
        int error;

        if ((error = pci_request_region (pdev, bars [cur_bar], name)) != 0)
        {
            printk (DEBUG_HEADER "Unable to reserve MMIO region (%d, %d), error %d.\n",
                    cur_bar, bars [cur_bar], error);

            /* Release previously reserved regions */

            while (--cur_bar > -1)
                pci_release_region (pdev, bars [cur_bar]);

            return 0;
        }
    }
#endif

    /* Now, map everything */

    for (cur_bar = 0; cur_bar < num_bars; cur_bar++)
    {
        unsigned long reg  = pci_resource_start (pdev, bars [cur_bar]);
        unsigned int  size = pci_resource_len (pdev, bars [cur_bar]);

        if ((maps [cur_bar] = ioremap (reg, size)) == NULL)
        {
            printk (DEBUG_HEADER "Mem region at %lx (size %u, %d [%d]) unavailable - already in use?\n",
                    reg, size, bars [cur_bar], cur_bar);

            /* Unmap and release previous maps */

            LinuxUnMapRegions (_pdev, bars, num_bars, maps);

            return 0;
        }
    }
        
    return 1;
}


int LinuxGetRegionSize (void *_pdev, int bar)
{
    struct pci_dev *pdev = (struct pci_dev *) _pdev;

    return pci_resource_len (pdev, bar);
}


void *LinuxMemMap (void *_pdev, char *name, unsigned long reg, u32 size)
{
    void *rez;
    struct pci_dev *pdev = (struct pci_dev *) _pdev;

    if (pdev)
        if (pci_request_regions (pdev, name))
        {
            printk (DEBUG_HEADER "Unable to reserve MMIO regions.\n");

            return NULL;
        }

    if ((rez = ioremap (reg, size)) == NULL)
    {
        printk (DEBUG_HEADER "Mem region at %p (size %d) unavailable - already in use?\n",
                (void *) reg, size);
    }

    return rez;
}


void LinuxMemUnmap (void *_pdev, void *reg)
{
    struct pci_dev *pdev = (struct pci_dev *) _pdev;

    if (pdev)
        pci_release_regions (pdev);

    iounmap (reg);
}


int LinuxClaimIrq (void *context, void *_pdev, char *name, void *isr)
{
    struct pci_dev *pdev = (struct pci_dev *) _pdev;

    ATTO_LOG (KERN_INFO, ATTO_LOG_INIT,
              "LinuxClaimIrq irq=%d (%p, %p, %s, %p)\n",
              pdev->irq, context, _pdev, name, isr);

    if (request_irq (pdev->irq,
                     isr,
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,18)
                     SA_INTERRUPT | SA_SHIRQ,
#else
                     IRQF_DISABLED | IRQF_SHARED,
#endif
                     name,
                     context))
    {
        printk (DEBUG_HEADER "Unable to request IRQ %02X.\n", pdev->irq);
        return 0;
    }

    return 1;
}

int LinuxRegisterPollHandler (void *context, void *_pdev, void *_shost, void *isr)
{
#ifdef ATTO_VMK_50
    struct pci_dev *pdev = (struct pci_dev *) _pdev;
    struct Scsi_Host *shost = (struct Scsi_Host *) _shost;

    vmklnx_scsi_register_poll_handler (shost, pdev->irq, isr, context);

#endif

    return 0;

}


void LinuxReleaseIrq (void *context, void *_pdev)
{
    struct pci_dev *pdev = (struct pci_dev *) _pdev;

    free_irq (pdev->irq, context);
}    


void LinuxPciUnmapSg (void *_cmd)
{
    Scsi_Cmnd *cmd = (Scsi_Cmnd *) _cmd;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,25)
    void *hostdata           = CMD_HOST(cmd)->hostdata;
    int nents                = cmd->use_sg;
    struct scatterlist *list = (struct scatterlist *) cmd->request_buffer;
    struct pci_dev *pdev     = GetPciDevFromHostData (hostdata);
    int dir                  = scsi_to_pci_dma_dir (cmd->sc_data_direction);

#ifdef ATTO_USE_KERNEL24_SCSI_TAPE_WORKAROUND
    /* See comment in atto_queuecommand for why this workaround is here. */

    int i;

    for (i = 0; i < nents; i++, list++)
        pci_unmap_single (pdev, list->dma_address, list->length, dir);

#else /* NOT ATTO_USE_KERNEL24_SCSI_TAPE_WORKAROUND */

    pci_unmap_sg (pdev, list, nents, dir);

#endif /* NOT ATTO_USE_KERNEL24_SCSI_TAPE_WORKAROUND */
#else /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,25) */
    scsi_dma_unmap (cmd);
#endif
}

unsigned long long LinuxPciMapSingle (void *_pdev, void *_cmd)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,25)
    /* This should never happen anymore with the new SCSI DMA stuff. */

    printk (DEBUG_HEADER "LinuxPciMapSingle called, that shouldn't happen!\n");

    return 0;
#else /* LINUX_VERSION_CODE < KERNEL_VERSION(2,6,25) */
    Scsi_Cmnd *cmd = (Scsi_Cmnd *) _cmd;
    struct pci_dev *pdev = (struct pci_dev *) _pdev;

    return pci_map_single (pdev,
                           cmd->request_buffer,
                           cmd->request_bufflen,
                           scsi_to_pci_dma_dir (cmd->sc_data_direction));
#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2,6,25) */
}

void LinuxPciUnmapSingle (void *_pdev, void *_cmd, unsigned long long bus_addr)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,25)
    printk (DEBUG_HEADER "LinuxPciUnmapSingle called, that shouldn't happen!\n");
#else /* LINUX_VERSION_CODE < KERNEL_VERSION(2,6,25) */
    struct pci_dev *pdev = (struct pci_dev *) _pdev;
    Scsi_Cmnd *cmd = (Scsi_Cmnd *) _cmd;
    pci_unmap_single (pdev,
                      bus_addr,
                      cmd->request_bufflen,
                      scsi_to_pci_dma_dir (cmd->sc_data_direction));
#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2,6,25) */
}


int LinuxStartTimer (u32 context, void *_timer, u32 millisecs, void (*fn)(unsigned long))
{
    struct timer_list **timer = (struct timer_list **) _timer;

    /* If this is the first call to this function, */
    /* we need to allocate some space for it.      */

    if (!(*timer))
        if (((*timer) = vmalloc (sizeof (struct timer_list))) == NULL)
        {
            printk (DEBUG_HEADER "Unable to allocate timer_list.\n");
            return 0;
        }

    init_timer ((*timer));

    (*timer)->function = fn;
    (*timer)->data     = context;
    (*timer)->expires  = jiffies + (HZ * millisecs) / 1000;

    add_timer ((*timer));

    return 1;
}


void LinuxKillTimer (void *_timer)
{
    struct timer_list **timer = (struct timer_list **) _timer;

    if (*timer)
    {
        del_timer_sync (*timer);
        vfree (*timer);
        (*timer) = NULL;
    }
}


void *LinuxAlloc (u32 size)
{
    void *rez;

    rez = vmalloc (size);

    if (rez)
        memset (rez, 0, size);

    return rez;
}


void LinuxFree (void *addr)
{
    if (addr)
        vfree (addr);
}


void *LinuxIoRemap (u32 offset, u32 size)
{
    return ioremap (offset, size);
}


void LinuxIoUnmap (void *addr)
{
    iounmap (addr);
}


unsigned long long LinuxGetCurrentTimeSecs (void)
{
    struct timeval Now;

    do_gettimeofday (&Now);

    return ((unsigned long long)(Now.tv_sec));
}


unsigned int LinuxGetCurrentTimeMs (void)
{
    struct timeval Now;

    do_gettimeofday (&Now);

    return Now.tv_sec * 1000L + (Now.tv_usec / 1000L);
}


unsigned long long LinuxGetCurrentTimeUs (void)
{
    struct timeval Now;

    do_gettimeofday (&Now);

    return (unsigned long long) Now.tv_sec * 1000000L + (unsigned long long) Now.tv_usec;
}


int LinuxWaitInit (void **_queue)
{
    wait_queue_head_t **queue = (wait_queue_head_t **) _queue;

    if (*queue)
    {
        printk (DEBUG_HEADER "LinuxWaitInit called twice for same wait_queue!\n");
        vfree (*queue);
    }

    if (((*queue) = vmalloc (sizeof (wait_queue_head_t))) == NULL)
        return 0;

    init_waitqueue_head (*queue);

    return 1;
}


int LinuxWaitOn (void **_queue, int *conditional)
{
    wait_queue_head_t **queue = (wait_queue_head_t **) _queue;

    if (!(*queue))
    {
        printk (DEBUG_HEADER "LinuxWaitOn called for invalid wait_queue!\n");
        return 0;
    }

    /* Sometimes wait_event_interruptible constantly returns an error          */
    /* (-ERESTARTSYS), ususally when closing ConfigTool while user commands    */
    /* are outstanding.  In this case we simply have to poll for completion.   */

    while (!(*conditional))
    {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11)
        wait_event_interruptible (**queue, (*conditional));
#else
        lock_kernel ();
        interruptible_sleep_on_timeout (*queue, 1);
        unlock_kernel ();
#endif
    };

    return 1;
}


void LinuxWaitWake (void **_queue)
{
    wait_queue_head_t **queue = (wait_queue_head_t **) _queue;

    if (!(*queue))
    {
        printk (DEBUG_HEADER "LinuxWaitWake called for invalid wait_queue!\n");
        return;
    }

    wake_up_interruptible (*queue);
}


void LinuxWaitFree (void **_queue)
{
    if (*_queue)
    {
        vfree (*_queue);
        (*_queue) = NULL;
    }
}

/* These get initialized along with the tasklets in LinuxDpcInit */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20) && !defined(ATTO_VMK_50)
static struct work_struct atto_workers [ATTO_MAX_DEVICES];
#else
static struct delayed_work atto_workers [ATTO_MAX_DEVICES];
#endif

#ifdef ATTO_RESCANS_ENABLED

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20) && !defined(ATTO_VMK_50)
void LinuxRescanHostRoutine (void *worker)
#else
void LinuxRescanHostRoutine (struct work_struct *worker)
#endif
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20) && !defined(ATTO_VMK_50)
    unsigned int index = (struct work_struct *) worker - atto_workers;
#else
    unsigned int index = (struct delayed_work *) worker - atto_workers;
#endif
    struct Scsi_Host *sh = (struct Scsi_Host *) OsGetAdapterHost (index);

    if (sh)
    {
        unsigned int scan_id = -1;
        unsigned char scan_state;

        ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TARG, &(sh->shost_gendev),
                  "LinuxRescanHostRoutine index:%d\n", index);

        if ((scan_id = OsGetNextScanId (index, &scan_state)) != -1)
        {
            int lun;
            struct scsi_device *scsi_dev;

            ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TARG, &(sh->shost_gendev),
                          "  scan_state:%d\n", scan_state);

#if defined (ATTO_USE_TRANSPORT) && defined (ATTO_TRANSPORT)
            if (use_transport_layer)
            {
#if defined (ATTO_RPORTS) && defined (CELERITYFC)
                if (scan_state == SCAN_STATE_LOGIN)
                {
                    celerityfc_add_rport(sh, scan_id);
                }
                else if (scan_state == SCAN_STATE_LOGOUT)
                    celerityfc_remove_rport(sh, scan_id);
                else if (scan_state == SCAN_STATE_LUN_CHANGE)
                    celerityfc_scan_rport(sh, scan_id);
#elif defined (ATTO_SAS_PORTS) && defined (ESASHBA)

                mutex_lock (&esas_topology[index].sas_topology_mutex);

                if (scan_state == SCAN_STATE_LOGIN)
                {
                    if (!esas_topology_add_device (&esas_topology[index],
                                                   scan_id))
                        printk (DEBUG_HEADER "Failed to add device at target 0x%04x.\n",
                                scan_id);
                }
                else if (scan_state == SCAN_STATE_LOGOUT)
                {
                    if (esas_topology_remove_device (&esas_topology[index],
                                                     scan_id))
                        printk (DEBUG_HEADER "Failed to remove device at target 0x%04x.\n",
                                scan_id);
                }
                else if (scan_state == SCAN_STATE_LUN_CHANGE)
                    esas_topology_scan_device (&esas_topology[index],
                                               scan_id);

                mutex_unlock (&esas_topology[index].sas_topology_mutex);

#endif /* defined (ATTO_SAS_PORTS) && defined (ESASHBA) */
                /* Reschedule this task to pick up any other notifications. */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20) && !defined(ATTO_VMK_50)
                schedule_work (&atto_workers [index]);
#else
                schedule_work (&(atto_workers [index].work));
#endif
                return;
            }
#endif /* ATTO_USE_TRANSPORT */

            if (scan_state == SCAN_STATE_LOGOUT
            ||  scan_state == SCAN_STATE_LUN_CHANGE)
            {
                for (lun = 0; lun < ATTO_MAX_LUN; lun++)
                {
                    scsi_dev = scsi_device_lookup (sh, 0, scan_id, lun);

                    if (scsi_dev)
                    {
#ifdef SDEV_OFFLINE
                        scsi_device_set_state (scsi_dev, SDEV_OFFLINE);
#endif
                        scsi_remove_device (scsi_dev);
                        scsi_device_put (scsi_dev);
                    }
                }
            }
            if (scan_state == SCAN_STATE_LOGIN
            ||  scan_state == SCAN_STATE_LUN_CHANGE)
            {
                for (lun = 0; lun < ATTO_MAX_LUN; lun++)
                {
                    scsi_dev = scsi_device_lookup (sh, 0, scan_id, lun);

                    if (scsi_dev)
                        scsi_device_put (scsi_dev);
                    else
                        scsi_add_device (sh, 0, scan_id, lun);
                }
            }

            /* Reschedule this task to pick up any other notifications. */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20) && !defined(ATTO_VMK_50)
            schedule_work (&atto_workers [index]);
#else
            schedule_work (&(atto_workers [index].work));
#endif
        }
    }
}


#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20) && !defined(ATTO_VMK_50)
void LinuxWorkRoutine (void *worker)
#else
void LinuxWorkRoutine (struct work_struct *worker)
#endif
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20) && !defined(ATTO_VMK_50)
    unsigned int index = (struct work_struct *) worker - atto_workers;
#else
    unsigned int index = (struct delayed_work *) worker - atto_workers;
#endif

#ifdef _MP_
    int ret;

    /* Multipathing code does additional work besides handling rescans. */
    /* It will call LinuxAddRemoveTarget as necessary. OsMpWorkRoutine  */
    /* will return -1 if there's nothing else to do, otherwise it       */
    /* returns the number of milliseconds to wait before calling again. */
    /* This helps in throttling commands during MP discovery/activation.*/

    ret = OsMpWorkRoutine (index);
    if (ret > 0)
        schedule_delayed_work (&atto_workers [index],
                               msecs_to_jiffies (ret));
    else if (ret == 0)
        schedule_delayed_work (&atto_workers [index],
                               10);

#else /* !_MP_ */


#ifdef ATTO_RESCANS_ENABLED
    /* Only reschedule if the core-side adapter is allowing it. */
    if ( OsWorkAllowed(index) )
        LinuxRescanHostRoutine (worker);
#endif /* ATTO_RESCANS_ENABLED */

#endif /* !_MP_ */
}


#ifdef _MP_
void LinuxAddRemoveTarget (void *_sh, unsigned int target_id, int lun, int remove)
{
    struct Scsi_Host *sh = _sh;
    struct scsi_device *scsi_dev = scsi_device_lookup (sh, 0, target_id, lun);

    if (scsi_dev)
    {
        if (!remove)
            printk (DEBUG_HEADER "Tried to add an existing SCSI device (%d:%d).\n",
                    target_id, lun);
        else
        {
#ifdef SDEV_OFFLINE
            scsi_device_set_state (scsi_dev, SDEV_OFFLINE);
#endif
            scsi_remove_device (scsi_dev);
        }

        scsi_device_put (scsi_dev);
    }
    else
    {
        if (remove)
        {
            printk (DEBUG_HEADER "Tried to remove a non-existent SCSI device (%d:%d).\n",
                    target_id, lun);
            return;
        }

        scsi_add_device (sh, 0, target_id, lun);
    }        
}
#endif /* _MP_ */

#endif /* ATTO_RESCANS_ENABLED */

void LinuxScheduleWork (unsigned int index)
{
#ifdef ATTO_RESCANS_ENABLED
    if (index >= ATTO_MAX_DEVICES)
    {
        printk (DEBUG_HEADER "LinuxScheduleWork tried to use an invalid index.\n");
        return;
    }

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20) && !defined(ATTO_VMK_50)
    schedule_work (&atto_workers [index]);
#else
    schedule_work (&(atto_workers [index].work));
#endif

#else
    printk (DEBUG_HEADER "LinuxScheduleWork called for 2.4 kernel.\n");
#endif
}


static struct tasklet_struct atto_tasklet [ATTO_MAX_DEVICES];

void *LinuxDpcInit (unsigned int index, void *callback)
{
    if (index >= ATTO_MAX_DEVICES)
    {
        printk (DEBUG_HEADER "LinuxDpcInit tried to initialize an invalid index.\n");
        return NULL;
    }

    memset (&atto_tasklet [index], 0, sizeof (struct tasklet_struct));

    tasklet_init (&atto_tasklet [index],
                  (void (*)(unsigned long)) callback,
                  (unsigned long) index);

#ifdef ATTO_RESCANS_ENABLED
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    memset (&atto_workers [index], 0, 
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20) && !defined(ATTO_VMK_50)
            sizeof (struct work_struct));
#else
            sizeof (struct delayed_work));
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20) && !defined(ATTO_VMK_50)
    INIT_WORK (&atto_workers [index],
               LinuxWorkRoutine,
               &atto_workers [index]);
#else /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20) */
    INIT_DELAYED_WORK (&atto_workers [index], LinuxWorkRoutine);
#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20) */
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0) */
#endif /* ATTO_RESCANS_ENABLED */

    return &atto_tasklet [index];
}


void LinuxDpcQueue (void *dpc_context)
{
    tasklet_hi_schedule (dpc_context);
}

void LinuxDpcStop (unsigned int index)
{
    if (index >= ATTO_MAX_DEVICES)
    {
        printk (DEBUG_HEADER "LinuxDpcInit tried to initialize an invalid index.\n");
        return;
    }

#ifdef ATTO_RESCANS_ENABLED
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    /* Cancel any outstanding work that hasn't started, then flush */
    /* the shared work queue to synchronize the worker thread.     */
    cancel_delayed_work (&atto_workers [index]);
    flush_scheduled_work ();
#endif /* LINUX_VERSION_CODE >= KRENEL_VERSION(2,6,0) */
#endif /* ATTO_RESCANS_ENABLED */

    return;
}


void LinuxDisableIrq (unsigned int irq)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
    disable_irq (irq);
#endif
}


void LinuxEnableIrq (unsigned int irq)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
    enable_irq (irq);
#endif
}


void LinuxUnregisterScsiHost (void *_host)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    struct Scsi_Host *host = _host;

    if (host)
    {
#if defined (CELERITYFC) && defined (ATTO_USE_TRANSPORT) && LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,12)
        if (use_transport_layer)
        {
            ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_XPORT, &(host->shost_gendev),
                          "fc_remove_host() called\n");
            fc_remove_host (host);
        }
#elif defined (ESASHBA) && defined (ATTO_USE_TRANSPORT) && LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
        if (use_transport_layer)
        {
            ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_XPORT, &(host->shost_gendev),
                          "sas_remove_host() called\n");
            sas_remove_host (host);
        }
#endif
        ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_SCSI, &(host->shost_gendev),
                      "scsi_remove_host() called\n");
        scsi_remove_host (host);
        ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_SCSI, &(host->shost_gendev),
                      "scsi_host_put() called\n");
        scsi_host_put (host);
    }
#endif
}


void LinuxUnregisterPci (void *_pcid)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    struct pci_dev *pcid = (struct pci_dev *) _pcid;

    if (pcid)
    {
        ATTO_LOG (KERN_INFO, ATTO_LOG_PCI,
                  "pci_set_drv_data(%p, NULL) called\n", pcid);
        pci_set_drvdata (pcid, NULL);
    }
#endif
}

void LinuxResetDetected (void *_sh, int channel)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    struct Scsi_Host *sh = (struct Scsi_Host *) _sh;

    if (sh)
    {
        ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_SCSI, &(sh->shost_gendev),
                      "scsi_report_bus_reset() called; channel %d\n", channel);
        scsi_report_bus_reset (sh, channel);
    }
#endif
}


void LinuxAcquireSpinLock (void *_lock, unsigned long *_flags)
{
#if defined(CONFIG_SMP)
    spinlock_t *lock = (spinlock_t *) _lock;
    spin_lock_irqsave (lock, *(_flags));
#endif
}


void LinuxReleaseSpinLock (void *_lock, unsigned long *_flags)
{
#if defined(CONFIG_SMP)
    spinlock_t *lock = (spinlock_t *) _lock;
    spin_unlock_irqrestore (lock, *(_flags));
#endif
}


int LinuxIsSpinLockHeld (void *_lock)
{
#if defined(CONFIG_SMP)
    spinlock_t *lock = (spinlock_t *) _lock;
    return spin_is_locked (lock);
#else
    return 0;
#endif
}


void *LinuxInitSpinLock (void)
{
#if defined(CONFIG_SMP)
    /* On non-SMP systems, sizeof (spinlock_t) is 0, and */
    /* vmalloc crashes when asked to allocate 0 bytes.   */

    spinlock_t *lock;

    if ((lock = (spinlock_t *) vmalloc (sizeof (*lock))) != NULL)
        spin_lock_init (lock);

    return lock;
#else
    /* Caller expects non-NULL return if successful.  Since this */
    /* value will never actually be used for anything on non-SMP */
    /* machines, just return a meaningless non-NULL value.       */

    return ((void *) 42);
#endif
}


void LinuxUninitSpinLock (void *lock)
{
#if defined(CONFIG_SMP)
    if (lock)
        vfree (lock);
#endif
}


int LinuxGetPageSize (void)
{
    return PAGE_SIZE;
}


void LinuxOutb (unsigned char v, unsigned long a)
{
    outb (v, a);
}


unsigned char LinuxInb (unsigned long a)
{
    return inb (a);
}


int LinuxSetDmaMask (void *_pdev, int size)
{
    struct pci_dev *pdev = (struct pci_dev *) _pdev;

    if (size == 32)
    {
        if (pci_set_dma_mask (pdev, (u64) 0xFFFFFFFFULL) == 0)
            return 1;
        else
        {
            printk (DEBUG_HEADER "Error setting 32-bit DMA mask!\n");
            return 0;
        }
    }

    if (size == 64)
    {
        if (pci_set_dma_mask (pdev, 0xFFFFFFFFFFFFFFFFULL) == 0)
        {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
            return 1;
#else
#ifdef CELERITYFC
            if (pci_set_consistent_dma_mask (pdev, 0xFFFFFFFFFFFFFFFFULL) == 0)
#else
            if (pci_set_consistent_dma_mask (pdev, (u64) 0xFFFFFFFFULL) == 0)
#endif
                return 1;
            else
            {
                printk (DEBUG_HEADER "Error setting 64-bit consistent DMA mask!\n");
                return 0;
            }
#endif
        }
        else
        {
            printk (DEBUG_HEADER "Error setting 64-bit DMA mask!\n");
            return 0;
        }
    }

    printk (DEBUG_HEADER "LinuxSetDmaMask invalid size %d!?\n", size);
    return 0;
}


char *LinuxGetKernelVersion (void)
{
    return UTS_RELEASE;
}


void LinuxYield (void)
{
#ifdef __VMKLNX__
    return;
#else
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
    /* Yield the CPU if we are doing something potentially */
    /* time-consuming.  However, this is not safe to do    */
    /* in certain contexts, because it will schedule().    */

    if (!in_atomic () && !irqs_disabled ())
        cond_resched ();
#endif
#endif
}


int LinuxGetErrno (int atto_errno)
{
    switch (atto_errno)
    {
    case ATTO_BUSY:           return -EBUSY;
    case ATTO_NO_DEVICE:      return -ENODEV;
    case ATTO_NOT_IMPLMNTD:   return -ENOSYS;
    case ATTO_INVALID_ARG:    return -EINVAL;
    case ATTO_OUT_OF_MEMORY:  return -ENOMEM;
    }

    return -ENOSYS;
}


#define POINTER_TO_DWORD(x)    ((unsigned int) (unsigned long) (x))

void LinuxExchangeOffsets (PSCSICMD_OFFSETS scsicmd_off,
                           int pci_dev_offset)
{
    Scsi_Cmnd *null_cmd = NULL;

    scsicmd_off->scsi_done_offset    = POINTER_TO_DWORD (&null_cmd->scsi_done);
    scsicmd_off->cmd_len_offset      = POINTER_TO_DWORD (&null_cmd->cmd_len);
    scsicmd_off->sense_buffer_offset = POINTER_TO_DWORD (&null_cmd->sense_buffer);
    scsicmd_off->data_dir_offset     = POINTER_TO_DWORD (&null_cmd->sc_data_direction);
    scsicmd_off->cmnd_offset         = POINTER_TO_DWORD (&null_cmd->cmnd);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,25)
    scsicmd_off->req_bufflen_offset  = POINTER_TO_DWORD (&null_cmd->request_bufflen);
    scsicmd_off->req_buffer_offset   = POINTER_TO_DWORD (&null_cmd->request_buffer);
    scsicmd_off->use_sg_offset       = POINTER_TO_DWORD (&null_cmd->use_sg);
#else
    scsicmd_off->req_bufflen_offset  = POINTER_TO_DWORD (&null_cmd->sdb.length);
    scsicmd_off->req_buffer_offset   = POINTER_TO_DWORD (&null_cmd->sdb.table.sgl);
    scsicmd_off->use_sg_offset       = POINTER_TO_DWORD (&null_cmd->sdb.table.nents);
    scsicmd_off->flags               = SCSICMD_FLAGS_USE_SENSE_PTR;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,25)
    scsicmd_off->flags              |= SCSICMD_FLAGS_USE_CMD_PTR;
#endif
#endif
    scsicmd_off->sense_buffer_size   = SCSI_SENSE_BUFFERSIZE;
    scsicmd_off->data_dir_unknown    = SCSI_DATA_UNKNOWN;
    scsicmd_off->data_dir_write      = SCSI_DATA_WRITE;
    scsicmd_off->data_dir_read       = SCSI_DATA_READ;
    scsicmd_off->data_dir_none       = SCSI_DATA_NONE;
    scsicmd_off->tag_value_head      = HEAD_OF_QUEUE_TAG;
    scsicmd_off->tag_value_simple    = SIMPLE_QUEUE_TAG;
    scsicmd_off->tag_value_ordered   = ORDERED_QUEUE_TAG; 

    scsicmd_off->error_success       = DID_OK;
    scsicmd_off->error_abort         = DID_ABORT;
    scsicmd_off->error_error         = DID_ERROR;
    scsicmd_off->error_no_connect    = DID_NO_CONNECT;
    scsicmd_off->error_reset         = DID_RESET;
    scsicmd_off->error_busy          = DID_BUS_BUSY;

    global_pci_dev_offset = pci_dev_offset;
}

/* The do_div macro is NOT defined in the vmklinux API. It is currently    */
/* only used for performance tracking/reporting in the core. However, this */
/* method must not be used when building vmware drivers!                   */

void LinuxDivideQwords (unsigned long long *Numerator,
                        unsigned int       *Denominator,
                        unsigned long long *Result)
{
    unsigned long long tmp = (*Numerator);
    do_div (tmp, (*Denominator));
    (*Result) = tmp;
}


void *LinuxGetReqBuffer (void *_cmd, unsigned int *BuffLength)
{
    Scsi_Cmnd *cmd = (Scsi_Cmnd *) _cmd;


    struct scatterlist *sgl = 
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23)
            (struct scatterlist *) cmd->request_buffer;

    if (cmd->use_sg)
#else
            (struct scatterlist *) scsi_sglist (cmd);

    if (scsi_sg_count (cmd))
#endif
    {
        if (((*BuffLength) = sg_dma_len (&sgl [0])) == 0)
            return NULL;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
#ifdef ATTO_VMK_50
        return kmap_atomic (sg_page (&sgl[0]), KM_IRQ0) + vmklnx_sg_offset (&sgl[0]);
#else
        return kmap_atomic (sgl [0].page, KM_IRQ0) + sgl [0].offset;
#endif /* ATTO_VMK_50 */
#else
        return kmap_atomic (sg_page (&sgl [0]), KM_IRQ0) + sgl [0].offset;
#endif
    }

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
    if (((*BuffLength) = cmd->request_bufflen) == 0)
#else
    if (((*BuffLength) = scsi_bufflen (cmd)) == 0)
#endif
        return NULL;

    return (void *) sgl;
}


void LinuxReleaseReqBuffer (void *_cmd, void *_buf)
{
    Scsi_Cmnd *cmd;
    unsigned char *buf;
    struct scatterlist *sgl;

    /* We don't just assign these values when we declare them because */
    /* kunmap_atomic is a NOP on some kernels (x64?), so assigning    */
    /* them up there leads to spurious "unused variable" warnings.    */

    cmd = (Scsi_Cmnd *) _cmd;
    buf = (unsigned char *) _buf;

    sgl = 
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23)
            (struct scatterlist *) cmd->request_buffer;

    if (cmd->use_sg)
#else
            (struct scatterlist *) scsi_sglist (cmd);

    if (scsi_sg_count (cmd))
#endif
        kunmap_atomic (buf - sgl [0].offset, KM_IRQ0);
}


int LinuxInRecovery (void *_sh)
{
    struct Scsi_Host *sh = (struct Scsi_Host *) _sh;

    return (sh->shost_state == SHOST_RECOVERY
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,14)
        ||  sh->shost_state == SHOST_CANCEL_RECOVERY
        ||  sh->shost_state == SHOST_DEL_RECOVERY
#endif
            );
}


#if defined (ESASRAID) || defined (ESASRAID2) || defined (_MP_)
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
extern attocfg_fire_async_event fire_async_event;

void LinuxFireEvent (void *data, int len)
{
    if (!fire_async_event)
    {
        printk (DEBUG_HEADER "Error sending event message - handler not registered.\n");
        return;
    }

    if (!(*fire_async_event) (data, len))
        printk (DEBUG_HEADER "Error sending event message.\n");
}
#else

void LinuxFireEvent (void *data, int len)
{
    printk (DEBUG_HEADER "VDA event occurred, but not supported on this kernel.\n");
}

#endif
#endif /* ESASRAID(2) || _MP_*/

#ifdef ESASHBA
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
extern attocfg_fire_async_event fire_bb_event;

void LinuxFireBbEvent (void *data, int len)
{
    if (!fire_bb_event)
    {
        printk (DEBUG_HEADER "Error sending event message - handler not registered.\n");
        return;
    }

    if (!(*fire_bb_event) (data, len))
        printk (DEBUG_HEADER "Error sending event message.\n");
}
#else

void LinuxFireBbEvent (void *data, int len)
{
    printk (DEBUG_HEADER "BB event occurred, but not supported on this kernel.\n");
}

#endif
#endif /* ESASHBA */

#if defined(ESASRAID) || defined(ESASRAID2) \
 || (defined(ESASHBA) && !defined(ESAS2HBA))
void *LinuxPciFindDevice (unsigned int vendor, unsigned int device, void *_prev)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,25)
    return (void *) pci_get_device  (vendor, device, (struct pci_dev *) _prev);
#else
    return (void *) pci_find_device (vendor, device, (struct pci_dev *) _prev);
#endif
}

void LinuxPciReleaseDevice (void *_pdev)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,25)
    struct pci_dev *pdev = (struct pci_dev *) _pdev;

    pci_dev_put (pdev);
#endif
}

void *LinuxGetBridgePdev (void *_pdev)
{
    struct pci_dev *pdev = (struct pci_dev *) _pdev;
    struct pci_bus *bus = NULL;
    struct pci_dev *bus_pdev = NULL;

    if ((pdev == NULL)
    || ((bus = pdev->bus) == NULL)
    || ((bus_pdev = bus->self) == NULL))
    {
        printk (DEBUG_HEADER "Error accessing bridge config space (%p, %p, %p)\n",
                pdev, bus, bus_pdev);
        return NULL;
    }

    return bus_pdev;
}
#endif

void LinuxVerifyMemoryForIO (unsigned long long addr, unsigned int len)
{
}

void LinuxScsiStateChange (void *_host)
{
}

void LinuxScsiUnregisterHost (void *_host)
{
#ifdef __VMKLNX__
    return;
#else
    struct Scsi_Host *host = (struct Scsi_Host *) _host;

    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_SCSI, &(host->shost_gendev),
                  "scsi_unregister() called\n");
    scsi_unregister (host);
#endif
}

/* Generic implementation of the Linux completion api. */

void *LinuxCompletionAlloc(void)
{
#ifdef ATTO_COMPLETION
    return vmalloc (sizeof (struct completion));
#else
    printk (DEBUG_HEADER "Warning: Linux Completion API not available.");
    return NULL;
#endif
}

void LinuxCompletionFree (void *_cmpl)
{
#ifdef ATTO_COMPLETION
    if (_cmpl != NULL)
        vfree (_cmpl);
#else
    printk (DEBUG_HEADER "Warning: Linux Completion API not available.");
#endif
    return;
}

void LinuxCompletionInit (void *_cmpl)
{
#ifdef ATTO_COMPLETION
    struct completion *cmpl = (struct completion *)_cmpl;

    init_completion (cmpl);
#else
    printk (DEBUG_HEADER "Warning: Linux Completion API not available.");
#endif
    return;
}

void LinuxCompletionWait (void *_cmpl)
{
#ifdef ATTO_COMPLETION
    struct completion *cmpl = (struct completion *)_cmpl;

    wait_for_completion (cmpl);
#else
    printk (DEBUG_HEADER "Warning: Linux Completion API not available.");
#endif
    return;
}

void LinuxCompletionDone (void *_cmpl)
{
#ifdef ATTO_COMPLETION
    struct completion *cmpl = (struct completion *)_cmpl;

    complete (cmpl);
#else
    printk (DEBUG_HEADER "Warning: Linux Completion API not available.");
#endif
    return;
}

unsigned long LinuxCompletionWaitTimeout (void *_cmpl, unsigned long timeout /* msecs */)
{
    unsigned long timeleft = timeout;
#ifdef ATTO_COMPLETION
    struct completion *cmpl = (struct completion *)_cmpl;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11)
    timeleft = wait_for_completion_timeout (cmpl, msecs_to_jiffies (timeout));
#else
    printk (DEBUG_HEADER "wait_for_completion_timeout not implemented; ignoring timeout value.");
    wait_for_completion (cmpl);
#endif 

#else
    printk (DEBUG_HEADER "Warning: Linux Completion API not available.");
#endif
    return timeleft;
}

unsigned long LinuxCompletionWaitTimeoutInterruptible (void *_cmpl, unsigned long timeout /* msecs */)
{
    unsigned long timeleft = timeout;
#ifdef ATTO_COMPLETION
    struct completion *cmpl = (struct completion *)_cmpl;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11)
    timeleft = wait_for_completion_interruptible_timeout (cmpl, msecs_to_jiffies (timeout));
#else
    printk (DEBUG_HEADER "wait_for_completion_timeout_interruptible not implemented; ignoring timeout value.");
    wait_for_completion (cmpl);
#endif 

#else
    printk (DEBUG_HEADER "Warning: Linux Completion API not available.");
#endif
    return timeleft;
}


#ifdef __VMKLNX__
/* The VMkernel extends scsi_device to contain a dump_active field. */
/* This field is set when the VMkernel is dumping core. We need to  */
/* know this so that commands get completed during the dump phase.  */
int LinuxCmdGetDumpActive (void *_cmd)
{
    struct scsi_cmnd *cmd = (struct scsi_cmnd *)_cmd;
    struct scsi_device *sdev;

    if (cmd == NULL)
        return 0;

    sdev = cmd->device;
    if (sdev == NULL)
        return 0;

    return sdev->dumpActive;
}

#endif
void *LinuxGetHostDataFromScsiDev (void *_scsi_dev, int *target_id)
{
    Scsi_Device *sd = (Scsi_Device *) _scsi_dev;

    if (target_id)
        (*target_id) = sd->id;

    return sd->host->hostdata;
}

#ifdef ATTO_USE_NPIV

void LinuxSetVPortState (void *_shost, int newState)
{
    enum fc_port_state state;
    struct Scsi_Host *shost = (struct Scsi_Host *) _shost;
    struct fc_vport *vport;

    if (!shost)
    {
        ATTO_LOG (KERN_WARNING, ATTO_LOG_NPIV,
                  "LinuxSetVPortState called with NULL host\n");
        return;
    }

    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_NPIV, &(shost->shost_gendev),
                  "LinuxSetVPortState newstate:%d\n", newState);


    vport = dev_to_vport (shost->shost_gendev.parent);
    if (!vport)
    {
        ATTO_DEV_LOG (KERN_WARNING, ATTO_LOG_NPIV,
                      "LinuxSetVPortState called with NULL vport\n");
        return;
    }

    switch (newState)
    {
    case ATTO_NPIV_STATE_UNKNOWN:
        state = FC_VPORT_UNKNOWN;
        break;
    case ATTO_NPIV_STATE_ACTIVE:
        state = FC_VPORT_ACTIVE;
        break;
    case ATTO_NPIV_STATE_DISABLED:
        state = FC_VPORT_DISABLED;
        break;
    case ATTO_NPIV_STATE_LINKDOWN:
        state = FC_VPORT_LINKDOWN;
        break;
    case ATTO_NPIV_STATE_INITIALIZING:
        state = FC_VPORT_INITIALIZING;
        break;
    case ATTO_NPIV_STATE_NO_FABRIC:
        state = FC_VPORT_NO_FABRIC_SUPP;
        break;
    case ATTO_NPIV_STATE_FAB_RSRC:
        state = FC_VPORT_NO_FABRIC_RSCS;
        break;
    case ATTO_NPIV_STATE_LOGGINGOUT:
        state = FC_VPORT_FABRIC_LOGOUT;
        break;
    case ATTO_NPIV_STATE_REJECT_WWN:
        state = FC_VPORT_FABRIC_REJ_WWN;
        break;
    case ATTO_NPIV_STATE_FAILED:
        state = FC_VPORT_FAILED;
        break;
    default:
        state = FC_VPORT_UNKNOWN;
        break;
    }

    if (state != vport->vport_state)
    {
        vport->vport_last_state = vport->vport_state;
        vport->vport_state = state;
    }
}
#endif /* ATTO_USE_NPIV */
